from datetime import datetime
from .listener import Event, SocketIO, Timer, Signal, contains
from .errors import AbortBranch
import select, signal, time, operator
try:
    import epoll
except ImportError:
    epoll = None

LISTEN_KQUEUE = 0
LISTEN_SELECT = 0#.001
LISTEN_POLL = 0#10
SLEEP_SEC = .0003
SLEEP_TURBO = None

def set_sleep(s):
    global SLEEP_SEC
    SLEEP_SEC = s

def set_turbo(s):
    global SLEEP_TURBO
    SLEEP_TURBO = s

def kbint(signals):
    if signal.SIGINT in signals:
        return True
    raise KeyboardInterrupt("You have not set a Keyboard Interrupt callback. To do so, use: 'rel.signal(2, your_callback_function)'.")

class Registrar(object):
    def __init__(self):
        self.events = {'read':{},'write':{}}
        self.timers = []
        self.addlist = []
        self.rmlist = []
        self.signals = {}
        self.tick = 0
        self.run_dispatch = False
        self.error_check = False

    def report(self):
        return {
            "timers": len(self.timers),
            "signals": len(list(self.signals.keys())),
            "reads": len(list(self.events["read"].keys())),
            "writes": len(list(self.events["write"].keys()))
        }

    def signal_add(self, sig):
        self.signals[sig.sig] = sig

    def signal_remove(self, sig):
        if sig in self.signals:
            del self.signals[sig]

    def init(self):
        for sig in self.signals:
            self.signals[sig].reset()
        self.__init__()

    def event(self,callback,arg,evtype,handle):
        return Event(self,callback,arg,evtype,handle)

    def read(self,sock,cb,*args):
        return SocketIO(self,'read',sock,cb,*args)

    def write(self,sock,cb,*args):
        return SocketIO(self,'write',sock,cb,*args)

    def dispatch(self):
        self.run_dispatch = True
        while self.run_dispatch:
            if not self.loop():
                self.run_dispatch = False

    def loop(self):
        if SLEEP_TURBO and self.events["write"]:
            time.sleep(SLEEP_TURBO)
        else:
            time.sleep(SLEEP_SEC)
        self.tick = datetime.now().microsecond
        e = self.check_events()
        t = self.check_timers()
        return e or t or self.signals

    def abort(self):
        self.run_dispatch = False
        for ev_list in list(self.events.values()):
            for sockio in list(ev_list.values()):
                sockio.delete()

    def abort_branch(self):
        raise AbortBranch()

    def signal(self,sig,cb,*args):
        return Signal(self,sig,cb,*args)

    def timeout(self,delay,cb,*args):
        return Timer(self,delay,cb,*args)

    def add_timer(self, timer):
         self.addlist.append(timer)

    def remove_timer(self, timer):
        self.rmlist.append(timer)

    def check_timers(self):
        changes = len(self.addlist) or len(self.rmlist)
        for timer in self.addlist:
            if timer not in self.timers:
                self.timers.append(timer)
        self.addlist = []
        for timer in self.rmlist:
            if timer in self.timers:
                self.timers.remove(timer)
        self.rmlist = []
        changes and self.timers.sort(key=operator.attrgetter("expiration"))
        t = time.time()
        for timer in self.timers:
            if not timer.check(t):
                self.rmlist.append(timer)
        return bool(self.timers)

    def callback(self, etype, fd):
        try:
            self.events[etype][fd].callback()
        except AbortBranch as e:
            pass # just go on with other code :)

    def handle_error(self, fd):
        if fd in self.events['read']:
            self.callback('read', fd)
        if fd in self.events['write']:
            self.callback('write', fd)

class KqueueRegistrar(Registrar):
    def __init__(self):
        Registrar.__init__(self)
        if not hasattr(select, "kqueue"):
            raise ImportError("could not find kqueue -- need Python 2.6+ on BSD (including OSX)!")
        self.kq = select.kqueue()
        self.kqf = {
            "read": select.KQ_FILTER_READ,
            "write": select.KQ_FILTER_WRITE
        }

    def abort(self):
        Registrar.abort(self)
        self.kq.close()
    
    def add(self, event):
        self.events[event.evtype][event.fd] = event
        self.kq.control([select.kevent(event.fd, self.kqf[event.evtype], select.KQ_EV_ADD)], 0)
    
    def remove(self, event):
        if event.fd in self.events[event.evtype]:
            del self.events[event.evtype][event.fd]
            try:
                self.kq.control([select.kevent(event.fd, self.kqf[event.evtype], select.KQ_EV_DELETE)], 0)
            except Exception as e:
                pass #connection probably closed
    
    def check_events(self):
        if self.events['read'] or self.events['write']:
            elist = self.kq.control(None, 1000, LISTEN_KQUEUE)
            for e in elist:
                if e.filter == self.kqf['read']:
                    self.callback('read', e.ident)
                elif e.filter == self.kqf['write']:
                    self.callback('write', e.ident)
                else:
                    self.handle_error(e.ident)
            return True
        return False

class SelectRegistrar(Registrar):
    def __init__(self):
        Registrar.__init__(self)

    def add(self, event):
        self.events[event.evtype][event.fd] = event

    def remove(self, event):
        if event.fd in self.events[event.evtype]:
            del self.events[event.evtype][event.fd]

    def check_events(self):
        if self.events['read'] or self.events['write']:
            rlist = list(self.events['read'].keys())
            wlist = list(self.events['write'].keys())
            try:
                r,w,e = select.select(rlist,wlist,rlist+wlist,LISTEN_SELECT)
            except select.error:
                return kbint(self.signals)
            for fd in r:
                self.callback('read', fd)
            for fd in w:
                self.callback('write', fd)
            for fd in e:
                self.handle_error(fd)
            return True
        return False

class PollRegistrar(Registrar):
    def __init__(self):
        Registrar.__init__(self)
        try:
            self.poll = select.poll()
        except AttributeError:
            # Probably a platform without poll support, such as Windows
            raise ImportError("could not import poll")

    def add(self, event):
        self.events[event.evtype][event.fd] = event
        self.register(event.fd)

    def remove(self, event):
        if event.fd in self.events[event.evtype]:
            del self.events[event.evtype][event.fd]
            self.poll.unregister(event.fd)
            self.register(event.fd)

    def check_events(self):
        if self.events['read'] or self.events['write']:
            try:
                items = self.poll.poll(LISTEN_POLL)
            except select.error:
                return kbint(self.signals)
            except epoll.error:
                return kbint(self.signals)
            for fd,etype in items:
                if contains(etype,select.POLLIN) and fd in self.events['read']:
                    self.callback('read', fd)
                if contains(etype,select.POLLOUT) and fd in self.events['write']:
                    self.callback('write', fd)
                if contains(etype,select.POLLERR) or contains(etype,select.POLLHUP):
                    self.handle_error(fd)
            return True
        return False

    def register(self, fd):
        mode = 0
        if fd in self.events['read']:
            mode = mode|select.POLLIN
        if fd in self.events['write']:
            mode = mode|select.POLLOUT
        if mode:
            try:
                self.poll.register(fd, mode)
            except epoll.error:
                pass

class EpollRegistrar(PollRegistrar):
    def __init__(self):
        if epoll is None:
            raise ImportError("could not import epoll")
        Registrar.__init__(self)
        self.poll = epoll.poll()



"""
Parser for socket.io protocol

This implementation is inspired from the package python-socketio
written by Miguel Grinberg and available under the MIT license at
https://github.com/miguelgrinberg/python-socketio
"""

from engineio_client.emitter import Emitter
import six
import re
import json
import functools

import logging
logger = logging.getLogger(__name__)

class Packet(object):
    CONNECT      = 0
    DISCONNECT   = 1
    EVENT        = 2
    ACK          = 3
    ERROR        = 4
    BINARY_EVENT = 5
    BINARY_ACK   = 6

    def __init__(self, type=None, data=None, namespace=None, id=None):
        self.type = type
        self.data = data
        self.namespace = namespace or '/'
        self.id = id

    @property
    def type_string(self):
        return {
            self.CONNECT:      'connect',
            self.DISCONNECT:   'disconnect',
            self.EVENT:        'event',
            self.ACK:          'ack',
            self.ERROR:        'error',
            self.BINARY_EVENT: 'binary_event',
            self.BINARY_ACK:   'binary_ack'
        }[self.type]

    def __str__(self):
        return ' - '.join([str(i) for i in [self.type_string, self.id, self.namespace, self.data] if i])

PATTERN = '^'
PATTERN += '([0-6])'         # type
PATTERN += '(?:(\d+)-)?'     # number of attachments (optional)
PATTERN += '(?:(/[^,]+),?)?' # namespace (optional)
PATTERN += '(\d*)'           # message id (optional)
PATTERN += '(.*)'            # data
PATTERN += '$'

class Parser(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.packet = None
        self.raw_data = None
        self.num_attachments = 0
        self.attachments = []

    def decode(self, bytes):
        if not self.packet:
            packet_type, num_attachments, namespace, packet_id, data = self.decode_packet(bytes)
            self.packet = Packet(type=packet_type, namespace=namespace, id=packet_id)
            self.raw_data = data
            self.num_attachments = num_attachments
        else:
            self.attachments.append(bytes)

        if self.num_attachments != len(self.attachments):
            return None

        packet = self.packet
        packet.data = self.construct_data(self.raw_data, self.attachments)
        self.reset()
        return packet

    def decode_packet(self, bytes):
        matches = re.findall(PATTERN, bytes)
        if not matches:
            raise ParserException("Decoded packet is invalid: %s" % repr(bytes))

        items = matches[0]
        packet_type     = int(items[0])
        num_attachments = int(items[1]) if items[1] else 0
        namespace       = items[2]
        packet_id       = int(items[3]) if items[3] else None
        data            = json.loads(items[4]) if items[4] else None

        return packet_type, num_attachments, namespace, packet_id, data

    def construct_data(self, data, attachments):
        ret = data
        if isinstance(data, list):
            ret = [self.construct_data(item, attachments) for item in data]
        elif isinstance(data, dict):
            if data.get('_placeholder', False) and 0 <= data.get('num', -1) < len(attachments):
                ret = bytearray(attachments[data['num']])
            else:
                ret = {key: self.construct_data(value, attachments) for key, value in six.iteritems(data)}
        return ret

    def encode(self, packet):
        bytes = six.text_type()
        data, attachments = self.deconstruct_data(packet.data)

        if attachments:
            bytes += six.text_type(len(attachments)) + '-'
            if packet.type == Packet.EVENT:
                packet.type = Packet.BINARY_EVENT
            elif packet.type == Packet.ACK:
                packet.type = Packet.BINARY_ACK

        bytes = six.text_type(packet.type) + bytes

        if packet.namespace and packet.namespace != '/':
            bytes += packet.namespace
            if packet.id or data:
                bytes += ','
        if packet.id is not None:
            bytes += six.text_type(packet.id)
        if data is not None:
            bytes += json.dumps(data, separators=(',', ':'))

        return [bytes] + attachments

    def deconstruct_data(self, data, attachments=None):
        if attachments is None:
            attachments = []

        ret = data
        if isinstance(data, bytearray):
            attachments.append(data)
            ret = {'_placeholder': True, 'num': len(attachments) - 1}
        elif isinstance(data, (tuple, list)):
            ret = [self.deconstruct_data(item, attachments)[0]
                   for item in data]
        elif isinstance(data, dict):
            ret = {key: self.deconstruct_data(value, attachments)[0]
                   for key, value in six.iteritems(data)}
        return ret, attachments

class ParserException(Exception):
    pass

class PacketException(Exception):
    pass




__author__ = 'Yaraslau Byshyk'
__author_email__ = 'dev@shyk.by'
__description__ = 'SIO is short IO that even shorter that ShortIO'
__license__ = 'WTFPL'
__title__ = 'sio'
__url__ = 'https://github.com/byshyk/sio'
__version__ = '0.1.0'



from .version import __version__
from .rel import override, supported_methods, initialize, set_sleep, set_turbo, safe_read, read, write, timeout, signal, event, dispatch, loop, report, abort, abort_branch, thread, tick, init, sys, EV_PERSIST, EV_READ, EV_SIGNAL, EV_TIMEOUT, EV_WRITE



from .emitter import Emitter
from .utils import format_long

import logging
logger = logging.getLogger(__name__)

class Transport(Emitter):
    """Abstract base class for transports implementations
    """
    def __init__(self, client, scheme, hostname, port, path, parser=None):
        super(Transport, self).__init__()
        self.client = client
        self.scheme = scheme
        self.hostname = hostname
        self.port = port
        self.path = path
        self.parser = parser

        self.state = 'closed'

    def open(self):
        if self.state == 'closed':
            self.state = 'opening'
            self.do_open()

    def close(self, send=True):
        if self.state in ['open', 'opening']:
            if send:
                self.do_close()
            self.handle_close()

    def send(self, packets):
        if self.state != 'open':
            raise TransportNotOpenException()

        self.do_send(packets)

    def handle_error(self, error):
        logger.warning("Transport error: %s", error)
        self.emit('error', error)

    def handle_open(self):
        logger.debug("Transport open")
        self.state = 'open'
        self.emit('open')

    def handle_close(self):
        logger.debug("Transport closed")
        self.state = 'closed'
        self.emit('close')

    def handle_packet(self, packet):
        logger.debug(format_long("Received packet: %s - %s", packet.type_string, repr(packet.data)))
        self.emit('packet', packet)

    def do_open(self):
        raise NotImplementedError()

    def do_close(self):
        raise NotImplementedError()

    def do_send(self, packets):
        raise NotImplementedError()

class TransportException(Exception):
    pass

class TransportNotOpenException(TransportException):
    pass

